@@ -23,7 +23,7 @@
  * version of this file under either the License or the GPL.
  */
 
-static const char *RCSID = "@(#) $Header: /cvsroot/aolserver/nsimap/nsimap.c,v 1.1.1.1 2003/05/08 21:46:36 scottg Exp $, compiled: " __DATE__ " " __TIME__;
+static const char *RCSID = "@(#) $Header: /cvsroot/aolserver/nsimap/nsimap.c,v 1.13 2004/10/13 21:21:36 seryakov Exp $, compiled: " __DATE__ " " __TIME__;
 
 /*
  * nsimap.c -- Interface to c-client Mail library
@@ -100,8 +100,8 @@ static const char *RCSID = "@(#) $Header: /cvsroot/aolserver/nsimap/nsimap.c,v 1
  *    ns_imap check #s
  *      performs internal mailbox checking
  *
- *    ns_imap headers #s msgno ?-array name?
- *    ns_imap header #s msgno hdrname
+ *    ns_imap headers #s msgno ?-array name? ?-flags flags?
+ *    ns_imap header #s msgno hdrname ?-flags flags?
  *      fetches the complete, unfiltered RFC 822 format header of the specified
  *      message as a text string and returns it as a Tcl list in the form
  *      { name value name value ... } suitable for using array set command.
@@ -141,7 +141,7 @@ static const char *RCSID = "@(#) $Header: /cvsroot/aolserver/nsimap/nsimap.c,v 1
  *      performs necessary base64/qprint decoding. Something similar
  *      to ns_return.
  *
- *    ns_imap struct #s msgno ?-flags flags?-array name?
+ *    ns_imap struct #s msgno ?-flags flags? -array name?
  *    ns_imap bodystruct #s msgno part ?-flags flags? ?-array name?
  *      fetches all the structured information
  *      (envelope, internal date, RFC 822 size, flags, and body structure) for
@@ -193,6 +193,13 @@ static const char *RCSID = "@(#) $Header: /cvsroot/aolserver/nsimap/nsimap.c,v 1
  *           mail/ noselect mail/text {noinferiors unmarked}
  *           mail/private {noinferiors unmarked} mail/trash {noinferiors unmarked}
  *
+ *    ns_imap search #s searchCriteria
+ *      performs mailbox search and returns found message ids. Supports
+ *      IMAP2 search criteria only.
+ *      Example:
+ *           ns_imap search 1 "FROM john"
+ *           ns_imap search 1 "SUBJECT Meeting"
+ *
  *    ns_imap subscribe #s mailbox
  *    ns_imap unsusbscribe #s mailbox
  *      adds/removes the given name to/from the subscription list
@@ -259,7 +266,7 @@ static const char *RCSID = "@(#) $Header: /cvsroot/aolserver/nsimap/nsimap.c,v 1
 #include "fs.h"
 #include <setjmp.h>
 
-#define VERSION  "3.1"
+#define VERSION  "3.2.3"
 
 typedef struct _mailSession {
    struct _mailSession *next,*prev;
@@ -301,7 +308,6 @@ static char *strStripHtml(char *str,char *tags[]);
 static char *strLower(char *str);
 static void mailGc(mailServer *server);
 static void mm_parseline(ENVELOPE *env,char *hdr,char *data,char *host);
-static void mm_freeenvelope(void **ptr);
 void mm_getquota(MAILSTREAM *stream,char *qroot,QUOTALIST *qlist);
 
 static Ns_Tls mailTls;
@@ -349,7 +355,8 @@ Ns_ModuleInit(char *server, char *module)
 #ifdef SSL
     ssl_onceonlyinit ();
 #endif
-
+    // UW Imap sends USR2
+    ns_signal(SIGUSR2,SIG_IGN);
     path = Ns_ConfigGetPath(server,module,NULL);
     serverPtr = ns_calloc(1,sizeof(mailServer));
     serverPtr->server = server;
@@ -416,7 +423,13 @@ static void freeSession(mailServer *server,mailSession *session,int lock)
     if(session->next) session->next->prev = session->prev;
     if(session == server->sessions) server->sessions = session->next;
     if(lock) Ns_MutexUnlock(&server->mailMutex);
-    if(session->stream) mail_close_full(session->stream,0);
+    if(session->stream) {
+#ifdef LOCAL
+      if(session->stream->dtb && !strcmp(session->stream->dtb->name,"imap"))
+        ((IMAPLOCAL *)session->stream->local)->byeseen = 1;
+#endif
+      mail_close_full(session->stream,0);
+    }
     Ns_SetFree(session->params);
     Ns_SetFree(session->headers);
     ns_free(session->error);
@@ -531,14 +544,14 @@ static Tcl_Obj *mailEnvelope(ENVELOPE *en,Tcl_Interp *interp,char *arrayName)
 {
     Tcl_Obj *list = Tcl_NewListObj(0,0);
 
-    if(en->remail) mailPair(interp,list,"Remail",en->remail,0,arrayName);
-    if(en->date) mailPair(interp,list,"Date",en->date,0,arrayName);
-    if(en->subject) mailPair(interp,list,"Subject",en->subject,0,arrayName);
-    if(en->in_reply_to) mailPair(interp,list,"In-Reply-To",en->in_reply_to,0,arrayName);
-    if(en->message_id) mailPair(interp,list,"Message-Id",en->message_id,0,arrayName);
-    if(en->newsgroups) mailPair(interp,list,"Newsgroups",en->newsgroups,0,arrayName);
-    if(en->followup_to) mailPair(interp,list,"Followup-To",en->followup_to,0,arrayName);
-    if(en->references) mailPair(interp,list,"References",en->references,0,arrayName);
+    if(en->remail) mailPair(interp,list,"Remail",en->remail,(unsigned)0,arrayName);
+    if(en->date) mailPair(interp,list,"Date",en->date,(unsigned)0,arrayName);
+    if(en->subject) mailPair(interp,list,"Subject",en->subject,(unsigned)0,arrayName);
+    if(en->in_reply_to) mailPair(interp,list,"In-Reply-To",en->in_reply_to,(unsigned)0,arrayName);
+    if(en->message_id) mailPair(interp,list,"Message-Id",en->message_id,(unsigned)0,arrayName);
+    if(en->newsgroups) mailPair(interp,list,"Newsgroups",en->newsgroups,(unsigned)0,arrayName);
+    if(en->followup_to) mailPair(interp,list,"Followup-To",en->followup_to,(unsigned)0,arrayName);
+    if(en->references) mailPair(interp,list,"References",en->references,(unsigned)0,arrayName);
     if(en->to) mailAddress(interp,list,"To",en->to,arrayName);
     if(en->from) mailAddress(interp,list,"From",en->from,arrayName);
     if(en->cc) mailAddress(interp,list,"Cc",en->cc,arrayName);
@@ -605,7 +618,16 @@ static Tcl_Obj *mailStruct(BODY *body,MESSAGECACHE *cache,Tcl_Interp *interp,cha
       if(cache->draft) *p++ = 'X';
       *p = 0;
       mailPair(interp,list,"flags",buf,0,arrayName);
-      mailPair(interp,list,"size",0,cache->rfc822_size,arrayName);
+      mailPair(interp,list,"size",0,(unsigned long)cache->rfc822_size,arrayName);
+      mailPair(interp,list,"internaldate.day",0,(unsigned long)cache->day,arrayName);
+      mailPair(interp,list,"internaldate.month",0,(unsigned long)cache->month,arrayName);
+      mailPair(interp,list,"internaldate.year",0,(unsigned long)(BASEYEAR+cache->year),arrayName);
+      mailPair(interp,list,"internaldate.hours",0,(unsigned long)cache->hours,arrayName);
+      mailPair(interp,list,"internaldate.minutes",0,(unsigned long)cache->minutes,arrayName);
+      mailPair(interp,list,"internaldate.seconds",0,(unsigned long)cache->seconds,arrayName);
+      mailPair(interp,list,"internaldate.zoccident",0,(unsigned long)cache->zoccident,arrayName);
+      mailPair(interp,list,"internaldate.zhours",0,(unsigned long)cache->zhours,arrayName);
+      mailPair(interp,list,"internaldate.zminutes",0,(unsigned long)cache->zminutes,arrayName);
     }
     mailPair(interp,list,"type",(char*)mailType(body->type),0,arrayName);
     switch(body->encoding) {
@@ -690,6 +712,9 @@ static void mailGc(mailServer *server)
       if(now - session->access_time > idle_timeout) {
         mailSession *next = session->next;
         Ns_Log(Notice,"ns_imap: GC: inactive session %d: %s",session->id,session->stream->mailbox);
+        Ns_TlsSet(&mailTls,session);
+        /* In case of panic we will be thrown here again */
+        if(setjmp(session->jmp)) continue;
         freeSession(server,session,0);
         session = next;
         continue;
@@ -699,11 +724,10 @@ static void mailGc(mailServer *server)
     Ns_MutexUnlock(&server->mailMutex);
 }
 
-int tclOption(int objc,Tcl_Obj *CONST objv[],int start,char *name,int single)
+int tclOption(int objc,Tcl_Obj *CONST objv[],int i,char *name,int single)
 {
-   int i ;
-   for(i = start;i < objc;i++) {
-     if(!strcmp(name,Tcl_GetStringFromObj(objv[i],0)))
+   for(;i < objc;i++) {
+     if(!strcmp(name,Tcl_GetString(objv[i])))
        return single ? i : i+1 < objc ? i+1 : -1;
    }
    return -1;
@@ -716,15 +740,17 @@ int tclOption(int objc,Tcl_Obj *CONST objv[],int start,char *name,int single)
 static int
 MailCmd(ClientData arg,Tcl_Interp *interp,int objc,Tcl_Obj *CONST objv[])
 {
-    unsigned int i;
-    mailSession *session;
+    int index;
+    unsigned int num;
+    mailSession *session = 0;
     mailServer *server = arg;
+    unsigned long msg,flags = 0;
 
     enum commands {
         cmdGc, cmdSessions, cmdDecode, cmdEncode, cmdParseDate, cmdStripHtml,
         cmdOpen,
         cmdClose, cmdReopen, cmdStatus, cmdHeaders, cmdText, cmdBody, cmdError,
-        cmdExpunge, cmdAppend, cmdCopy, cmdMove, cmdPing,
+        cmdExpunge, cmdAppend, cmdCopy, cmdMove, cmdPing, cmdSearch,
         cmdCheck, cmdStruct, cmdMCreate, cmdMDelete, cmdMRename, cmdDelete,
         cmdUndelete, cmdSetFlags, cmdClearFlags, cmdList, cmdLsub, cmdSubscribe,
         cmdUnsubscribe, cmdNmsgs, cmdNrecent, cmdSort, cmdBodyStruct, cmdUid,
@@ -735,7 +761,7 @@ MailCmd(ClientData arg,Tcl_Interp *interp,int objc,Tcl_Obj *CONST objv[])
         "gc", "sessions", "decode", "encode", "parsedate", "striphtml",
         "open",
         "close", "reopen", "status", "headers", "text", "body", "error",
-        "expunge", "append", "copy", "move", "ping",
+        "expunge", "append", "copy", "move", "ping", "search",
         "check", "struct", "m_create", "m_delete", "m_rename", "delete",
         "undelete", "setflags", "clearflags", "list", "lsub", "subscribe",
         "unsubscribe", "n_msgs", "n_recent", "sort", "bodystruct", "uid",
@@ -755,8 +781,8 @@ MailCmd(ClientData arg,Tcl_Interp *interp,int objc,Tcl_Obj *CONST objv[])
         Tcl_AppendResult(interp, "wrong # args: should be ns_imap command #s ?args ...?",0);
         return TCL_ERROR;
       }
-      if(Tcl_GetIntFromObj(interp,objv[2],&i) != TCL_OK) return TCL_ERROR;
-      if(!(session = getSession(server,i))) {
+      if(Tcl_GetIntFromObj(interp,objv[2],&num) != TCL_OK) return TCL_ERROR;
+      if(!(session = getSession(server,num))) {
         Tcl_AppendResult(interp,"Invalid or expired mail session",0);
         return TCL_ERROR;
       }
@@ -793,14 +819,13 @@ MailCmd(ClientData arg,Tcl_Interp *interp,int objc,Tcl_Obj *CONST objv[])
     }
     case cmdStripHtml: {
         char *data;
-        static char *tags[] = { "BODY", "FRAME", "FRAMESET", "STYLE",
-                                "HEAD", "TITLE", "META", "DIV", "OBJECT",
-                                "EMBED", "BASE", 0 };
+        static char *tags[] = { "BODY", "FRAME", "FRAMESET","HEAD", "TITLE",
+                                "META", "DIV", "OBJECT", "EMBED", "BASE", 0 };
         if(objc < 3) {
           Tcl_AppendResult(interp, "wrong # args: should be ns_imap ",sCmd[cmd]," text ?tags?",0);
           return TCL_ERROR;
         }
-        data = strStripHtml(Tcl_GetStringFromObj(objv[2],0),tags);
+        data = strStripHtml(Tcl_GetString(objv[2]),tags);
         Tcl_SetResult(interp,data,(Tcl_FreeProc*)ns_free);
     }
     case cmdParseDate: {
@@ -837,21 +862,21 @@ MailCmd(ClientData arg,Tcl_Interp *interp,int objc,Tcl_Obj *CONST objv[])
         data = (char *)Tcl_GetByteArrayFromObj(objv[3],(int*)&len);
         // Decode BASE64 encoded text */
         if(!strcmp(Tcl_GetStringFromObj(objv[2],0),"base64")) {
-          data = (char *)rfc822_base64(data,i=len,&len);
+          data = (char *)rfc822_base64(data,num=len,&len);
           Tcl_SetObjResult(interp,Tcl_NewByteArrayObj(data,(int)len));
           fs_give((void**)&data);
           return TCL_OK;
         } else
         // Convert a quoted-printable string to an 8-bit string */
         if(!strcmp(Tcl_GetStringFromObj(objv[2],0),"qprint")) {
-          data = (char *)rfc822_qprint(data,i=len,&len);
+          data = (char *)rfc822_qprint(data,num=len,&len);
           Tcl_SetObjResult(interp,Tcl_NewByteArrayObj(data,(int)len));
           fs_give((void**)&data);
           return TCL_OK;
         } else
         // Convert a UTF7 an 8-bit string */
         if(!strcmp(Tcl_GetStringFromObj(objv[2],0),"utf7")) {
-          if((data = utf7_decode(data,(int)i=len,(int*)&len))) {
+          if((data = utf7_decode(data,num=len,(int*)&len))) {
             Tcl_SetObjResult(interp,Tcl_NewByteArrayObj(data,(int)len));
             ns_free(data);
             return TCL_OK;
@@ -871,21 +896,21 @@ MailCmd(ClientData arg,Tcl_Interp *interp,int objc,Tcl_Obj *CONST objv[])
         data = (char *)Tcl_GetByteArrayFromObj(objv[3],(int*)&len);
         // Convert an 8bit string to a quoted printable string */
         if(!strcmp(Tcl_GetStringFromObj(objv[2],0),"qprint")) {
-          data = (char *)rfc822_8bit(data,i=len,&len);
+          data = (char *)rfc822_8bit(data,num=len,&len);
           Tcl_SetObjResult(interp,Tcl_NewStringObj(data,(int)len));
           fs_give((void**)&data);
           return TCL_OK;
         } else
         // Convert an 8bit string to a base64 string */
         if(!strcmp(Tcl_GetStringFromObj(objv[2],0),"base64")) {
-          data = (char *)rfc822_binary(data,i=len,&len);
+          data = (char *)rfc822_binary(data,num=len,&len);
           Tcl_SetObjResult(interp,Tcl_NewStringObj(data,(int)len));
           fs_give((void**)&data);
           return TCL_OK;
         } else
         // Convert a 8-bit string into UTF7 */
         if(!strcmp(Tcl_GetStringFromObj(objv[2],0),"utf7")) {
-          if((data = utf7_encode(data,(int)i=len,(int*)&len))) {
+          if((data = utf7_encode(data,num=len,(int*)&len))) {
             Tcl_SetObjResult(interp,Tcl_NewStringObj(data,(int)len));
             ns_free(data);
             return TCL_OK;
@@ -962,8 +987,8 @@ MailCmd(ClientData arg,Tcl_Interp *interp,int objc,Tcl_Obj *CONST objv[])
           return TCL_ERROR;
         }
         if(!session->params) session->params = Ns_SetCreate("params");
-        if((i = Ns_SetFind(session->params,Tcl_GetStringFromObj(objv[3],0))) > -1)
-          Ns_SetPutValue(session->params,(int)i,Tcl_GetStringFromObj(objv[4],0));
+        if((index = Ns_SetFind(session->params,Tcl_GetStringFromObj(objv[3],0))) > -1)
+          Ns_SetPutValue(session->params,index,Tcl_GetStringFromObj(objv[4],0));
         else
           Ns_SetPut(session->params,Tcl_GetStringFromObj(objv[3],0),Tcl_GetStringFromObj(objv[4],0));
         break;
@@ -1017,6 +1042,7 @@ MailCmd(ClientData arg,Tcl_Interp *interp,int objc,Tcl_Obj *CONST objv[])
         break;
     case cmdPing:
         // Check if the stream is still active
+        if(!session->stream) break;
         Tcl_SetObjResult(interp,Tcl_NewIntObj(mail_ping(session->stream)));
         break;
     case cmdCheck:
@@ -1030,37 +1056,50 @@ MailCmd(ClientData arg,Tcl_Interp *interp,int objc,Tcl_Obj *CONST objv[])
     case cmdHeaders: {
         // Read message headers
         char *array = 0;
-        unsigned long msg;
         if(objc < 4) {
-          Tcl_AppendResult(interp, "wrong # args: should be ns_imap ",sCmd[cmd]," #s msgno ?-array name?",0);
+          Tcl_AppendResult(interp, "wrong # args: should be ns_imap ",sCmd[cmd]," #s msgno ?-array name? ?-flags flags?",0);
           return TCL_ERROR;
         }
-        if(Tcl_GetLongFromObj(interp,objv[3],&msg) != TCL_OK ||
-           msg <= 0 || msg > session->stream->nmsgs) {
+        if(Tcl_GetLongFromObj(interp,objv[3],&msg) != TCL_OK) return TCL_ERROR;
+        if(objc > 4) {
+          if((index = tclOption(objc,objv,4,"-flags",0)) > 0)
+            mailFlags(Tcl_GetStringFromObj(objv[index],0),&flags);
+        }
+        if(flags & FT_UID) {
+          if((msg = mail_msgno(session->stream,msg))) flags &= ~FT_UID;
+        }
+        if(msg <= 0 || msg > session->stream->nmsgs) {
           Tcl_AppendResult(interp,"Invalid message number",0);
           return TCL_ERROR;
         }
         if(mailHeaders(session,msg)) return TCL_ERROR;
         if(objc > 4) {
-          if((i = tclOption(objc,objv,4,"-array",0)) > 0)
-            array = Tcl_GetStringFromObj(objv[i],0);
+          if((index = tclOption(objc,objv,4,"-array",0)) > 0)
+            array = Tcl_GetStringFromObj(objv[index],0);
         }
         session->list = Tcl_NewListObj(0,0);
-        for(i = 0;i < Ns_SetSize(session->headers);i++)
-          mailPair(interp,session->list,Ns_SetKey(session->headers,i),Ns_SetValue(session->headers,i),0,array);
+        for(num = 0;num < Ns_SetSize(session->headers);num++)
+          mailPair(interp,session->list,Ns_SetKey(session->headers,num),
+                                        Ns_SetValue(session->headers,num),0,array);
         Tcl_SetObjResult(interp,session->list);
         break;
     }
     case cmdHeader: {
         // Returns specified header value
-        unsigned long msg;
         char *hdr;
         if(objc < 5) {
-          Tcl_AppendResult(interp, "wrong # args: should be ns_imap ",sCmd[cmd]," #s msgno hdrname",0);
+          Tcl_AppendResult(interp, "wrong # args: should be ns_imap ",sCmd[cmd]," #s msgno hdrname ?-flags flags?",0);
           return TCL_ERROR;
         }
-        if(Tcl_GetLongFromObj(interp,objv[3],&msg) != TCL_OK ||
-           msg <= 0 || msg > session->stream->nmsgs) {
+        if(Tcl_GetLongFromObj(interp,objv[3],&msg) != TCL_OK) return TCL_ERROR;
+        if(objc > 5) {
+          if((index = tclOption(objc,objv,5,"-flags",0)) > 0)
+            mailFlags(Tcl_GetStringFromObj(objv[index],0),&flags);
+        }
+        if(flags & FT_UID) {
+          if((msg = mail_msgno(session->stream,msg))) flags &= ~FT_UID;
+        }
+        if(msg <= 0 || msg > session->stream->nmsgs) {
           Tcl_AppendResult(interp,"Invalid message number",0);
           return TCL_ERROR;
         }
@@ -1072,18 +1111,23 @@ MailCmd(ClientData arg,Tcl_Interp *interp,int objc,Tcl_Obj *CONST objv[])
     case cmdText: {
         // Read the message text
         char *text;
-        unsigned long msg,len,flags = 0;
+        unsigned long len;
         if(objc < 4) {
           Tcl_AppendResult(interp, "wrong # args: should be ns_imap ",sCmd[cmd]," #s msgno ?-flags flags?",0);
           return TCL_ERROR;
         }
-        if(Tcl_GetLongFromObj(interp,objv[3],&msg) != TCL_OK ||
-           msg <= 0 || msg > session->stream->nmsgs) {
+        if(Tcl_GetLongFromObj(interp,objv[3],&msg) != TCL_OK) return TCL_ERROR;
+        if(objc > 4) {
+          if((index = tclOption(objc,objv,4,"-flags",0)) > 0)
+            mailFlags(Tcl_GetStringFromObj(objv[index],0),&flags);
+        }
+        if(flags & FT_UID) {
+          if((msg = mail_msgno(session->stream,msg))) flags &= ~FT_UID;
+        }
+        if(msg <= 0 || msg > session->stream->nmsgs) {
           Tcl_AppendResult(interp,"Invalid message number",0);
           return TCL_ERROR;
         }
-        if(objc > 4 && (i = tclOption(objc,objv,4,"-flags",0)))
-          mailFlags(Tcl_GetStringFromObj(objv[i],0),&flags);
         text = mail_fetchtext_full(session->stream,msg,&len,(int)flags);
         if(session->error) {
           Tcl_AppendResult(interp,session->error,0);
@@ -1094,30 +1138,35 @@ MailCmd(ClientData arg,Tcl_Interp *interp,int objc,Tcl_Obj *CONST objv[])
     }
     case cmdBody: {
         // Read the body part
-        BODY *body;
+        BODY *body = 0;
         PARAMETER *filename;
-        char *text,*fname,*data = 0;
-        unsigned long msg,len,flags = 0,decode = 0,mode = 0;
+        char *text,*fname = 0,*data = 0;
+        unsigned long len,decode = 0,mode = 0;
 
         if(objc < 5) {
           Tcl_AppendResult(interp, "wrong # args: should be ns_imap ",sCmd[cmd]," #s msgno part ?-flags flags? ?-decode? ?-file name? ?-return?",0);
           return TCL_ERROR;
         }
-        if(Tcl_GetLongFromObj(interp,objv[3],&msg) != TCL_OK ||
-           msg <= 0 || msg > session->stream->nmsgs) {
+        if(Tcl_GetLongFromObj(interp,objv[3],&msg) != TCL_OK) return TCL_ERROR;
+        if(objc > 5) {
+          if((index = tclOption(objc,objv,5,"-flags",0)) > 0)
+            mailFlags(Tcl_GetStringFromObj(objv[index],0),&flags);
+          if((index = tclOption(objc,objv,5,"-decode",1)) > 0)
+            decode = 1;
+          if((index = tclOption(objc,objv,5,"-return",1)) > 0)
+            mode = 1;
+          if((index = tclOption(objc,objv,5,"-file",0)) > 0) {
+            mode = 2;
+            fname = Tcl_GetStringFromObj(objv[index],0);
+          }
+        }
+        if(flags & FT_UID) {
+          if((msg = mail_msgno(session->stream,msg))) flags &= ~FT_UID;
+        }
+        if(msg <= 0 || msg > session->stream->nmsgs) {
           Tcl_AppendResult(interp,"Invalid message number",0);
           return TCL_ERROR;
         }
-        if(objc > 5) {
-          if((i = tclOption(objc,objv,5,"-flags",0)) > 0)
-            mailFlags(Tcl_GetStringFromObj(objv[i],0),&flags);
-          if((i = tclOption(objc,objv,5,"-decode",1)) > 0)
-            decode = 1;
-          if((i = tclOption(objc,objv,5,"-return",1)) > 0)
-            mode = 1;
-          if((i = tclOption(objc,objv,5,"-file",0)) > 0)
-            mode = 2,fname = Tcl_GetStringFromObj(objv[i],0);
-        }
         text = mail_fetchbody_full(session->stream,msg,Tcl_GetStringFromObj(objv[4],0),&len,(int)flags);
         if(text) body = mail_body(session->stream,msg,Tcl_GetStringFromObj(objv[4],0));
         if(!text || !body || session->error) {
@@ -1127,10 +1176,10 @@ MailCmd(ClientData arg,Tcl_Interp *interp,int objc,Tcl_Obj *CONST objv[])
         if(decode || mode) {
           switch(body->encoding) {
            case ENCBASE64:
-               text = data = (char *)rfc822_base64(text,i=len,&len);
+               text = data = (char *)rfc822_base64(text,num=len,&len);
                break;
            case ENCQUOTEDPRINTABLE:
-               text = data = (char *)rfc822_qprint(text,i=len,&len);
+               text = data = (char *)rfc822_qprint(text,num=len,&len);
           }
         }
         // Discover attachement file name
@@ -1152,13 +1201,14 @@ MailCmd(ClientData arg,Tcl_Interp *interp,int objc,Tcl_Obj *CONST objv[])
              break;
          }
          case 2:
+             Ns_Log(Debug,"nsimap: body: %s, %s",fname,filename?filename->value:0);
              if(*fname == 0 && filename) fname = filename->value;
-             if((i = open(fname,O_WRONLY|O_CREAT|O_TRUNC,0644)) == -1) {
+             if((index = open(fname,O_WRONLY|O_CREAT|O_TRUNC,0644)) == -1) {
                Tcl_AppendResult(interp,"Unable to create file:",fname,", ",strerror(errno),0);
                return TCL_ERROR;
              }
-             write((int)i,text,(int)len);
-             close((int)i);
+             write(index,text,(int)len);
+             close(index);
              break;
          default:
              Tcl_SetObjResult(interp,Tcl_NewByteArrayObj(text,(int)len));
@@ -1170,44 +1220,50 @@ MailCmd(ClientData arg,Tcl_Interp *interp,int objc,Tcl_Obj *CONST objv[])
         // Read the structure of a specified body section of a specific message
         BODY *body;
         char *array = 0;
-        unsigned long msg;
         if(objc < 5) {
-          Tcl_AppendResult(interp, "wrong # args: should be ns_imap ",sCmd[cmd]," #s msgno part ?-array name?",0);
+          Tcl_AppendResult(interp, "wrong # args: should be ns_imap ",sCmd[cmd]," #s msgno part ?-array name? ?-flags flags?",0);
           return TCL_ERROR;
         }
-        if(Tcl_GetLongFromObj(interp,objv[3],&msg) != TCL_OK ||
-           msg <= 0 || msg > session->stream->nmsgs) {
+        if(Tcl_GetLongFromObj(interp,objv[3],&msg) != TCL_OK) return TCL_ERROR;
+        if(objc > 5) {
+          if((index = tclOption(objc,objv,5,"-array",0)) > 0)
+            array = Tcl_GetStringFromObj(objv[index],0);
+          if((index = tclOption(objc,objv,5,"-flags",0)) > 0)
+            mailFlags(Tcl_GetStringFromObj(objv[index],0),&flags);
+        }
+        if(flags & FT_UID) {
+          if((msg = mail_msgno(session->stream,msg))) flags &= ~FT_UID;
+        }
+        if(msg <= 0 || msg > session->stream->nmsgs) {
           Tcl_AppendResult(interp,"Invalid message number",0);
           return TCL_ERROR;
         }
-        body = mail_body(session->stream,msg,Tcl_GetStringFromObj(objv[4],0));
-        if(objc > 5) {
-          if((i = tclOption(objc,objv,5,"-array",0)) > 0)
-            array = Tcl_GetStringFromObj(objv[i],0);
-        }
-        Tcl_SetObjResult(interp,mailStruct(body,0,interp,array));
+        if((body = mail_body(session->stream,msg,Tcl_GetStringFromObj(objv[4],0))))
+          Tcl_SetObjResult(interp,mailStruct(body,0,interp,array));
         break;
     }
     case cmdStruct: {
         // Read the whole message
         BODY *body;
         char *array = 0;
-        unsigned long msg,flags = 0;
         if(objc < 4) {
           Tcl_AppendResult(interp, "wrong # args: should be ns_imap ",sCmd[cmd]," #s msgno ?-flags flags? ?-array name?",0);
           return TCL_ERROR;
         }
-        if(Tcl_GetLongFromObj(interp,objv[3],&msg) != TCL_OK ||
-           msg <= 0 || msg > session->stream->nmsgs) {
-          Tcl_AppendResult(interp,"Invalid message number",0);
-          return TCL_ERROR;
-        }
+        if(Tcl_GetLongFromObj(interp,objv[3],&msg) != TCL_OK) return TCL_ERROR;
         /* Check for optional flags or/and array name */
         if(objc > 4) {
-          if((i = tclOption(objc,objv,4,"-flags",0)) > 0)
-            mailFlags(Tcl_GetStringFromObj(objv[i],0),&flags);
-          if((i = tclOption(objc,objv,4,"-array",0)) > 0)
-            array = Tcl_GetStringFromObj(objv[i],0);
+          if((index = tclOption(objc,objv,4,"-array",0)) > 0)
+            array = Tcl_GetStringFromObj(objv[index],0);
+          if((index = tclOption(objc,objv,4,"-flags",0)) > 0)
+            mailFlags(Tcl_GetStringFromObj(objv[index],0),&flags);
+        }
+        if(flags & FT_UID) {
+          if((msg = mail_msgno(session->stream,msg))) flags &= ~FT_UID;
+        }
+        if(msg <= 0 || msg > session->stream->nmsgs) {
+          Tcl_AppendResult(interp,"Invalid message number",0);
+          return TCL_ERROR;
         }
         if(!mail_fetchstructure_full(session->stream,msg,&body,(int)flags) || !body) {
           Tcl_AppendResult(interp,session->error,0);
@@ -1346,10 +1402,10 @@ MailCmd(ClientData arg,Tcl_Interp *interp,int objc,Tcl_Obj *CONST objv[])
           Tcl_AppendResult(interp, "wrong # args: should be ns_imap ",sCmd[cmd]," #s criteria reverse ?flags?",0);
           return TCL_ERROR;
         }
-        if(Tcl_GetIndexFromObj(interp,objv[3],sSort,"option",TCL_EXACT,&i) != TCL_OK) return TCL_ERROR;
+        if(Tcl_GetIndexFromObj(interp,objv[3],sSort,"option",TCL_EXACT,&num) != TCL_OK) return TCL_ERROR;
         memset(&pgm,0,sizeof(pgm));
-        pgm.function = iSort[i];
-        if(Tcl_GetIntFromObj(0,objv[4],&i) == TCL_OK) pgm.reverse = i ? 1 : 0;
+        pgm.function = iSort[num];
+        if(Tcl_GetIntFromObj(0,objv[4],&index) == TCL_OK) pgm.reverse = index ? 1 : 0;
         spg = mail_newsearchpgm();
         if(objc == 6) mailFlags(Tcl_GetStringFromObj(objv[5],0),&flags);
         if((ids = mail_sort(session->stream,0,spg,&pgm,flags))) {
@@ -1366,6 +1422,29 @@ MailCmd(ClientData arg,Tcl_Interp *interp,int objc,Tcl_Obj *CONST objv[])
         }
         break;
     }
+    case cmdSearch: {
+        // Search message headers
+	SEARCHPGM *crit = 0;
+        unsigned long id;
+        if(objc < 4) {
+          Tcl_AppendResult(interp, "wrong # args: should be ns_imap ",sCmd[cmd]," #s criteria",0);
+          return TCL_ERROR;
+        }
+        if(!session->stream->nmsgs) break;
+        if(!(crit = mail_criteria(Tcl_GetString(objv[3])))) break;
+        if(mail_search_full(session->stream,0,crit,SE_FREE)) {
+          session->list = Tcl_NewListObj(0,0);
+          for(id = 1;id <= session->stream->nmsgs;id++)
+            if(mail_elt(session->stream,id)->searched)
+              Tcl_ListObjAppendElement(interp,session->list,Tcl_NewIntObj(id));
+          Tcl_SetObjResult(interp,session->list);
+        }
+        if(session->error) {
+          Tcl_AppendResult(interp,session->error,0);
+          return TCL_ERROR;
+        }
+        break;
+    }
     case cmdList:
         // Read the list of mailboxes using optional pattern and substring
         if(objc < 5) {
@@ -1439,8 +1518,9 @@ MailCmd(ClientData arg,Tcl_Interp *interp,int objc,Tcl_Obj *CONST objv[])
           mail_status(session->stream,session->stream->mailbox,flags);
         }
         if(strstr(status_flags,"USERFLAGS")) {
-          for(i = 0; i < NUSERFLAGS && session->stream->user_flags[i]; ++i)
-            Tcl_ListObjAppendElement(interp,session->list,Tcl_NewStringObj(session->stream->user_flags[i],-1));
+          for(num = 0; num < NUSERFLAGS && session->stream->user_flags[num]; ++num)
+            Tcl_ListObjAppendElement(interp,session->list,
+                      Tcl_NewStringObj(session->stream->user_flags[num],-1));
         }
         Tcl_SetObjResult(interp,session->list);
         break;
@@ -1706,7 +1786,7 @@ static char *utf7_decode(char *in,int inlen,int *outlen)
     /* enforce end state */
     if(state != ST_NORMAL) return 0;
     /* allocate output buffer */
-    if((out = ns_malloc((*outlen) + 1)) == NULL) return 0;
+    if((out = ns_malloc((unsigned)(*outlen) + 1)) == NULL) return 0;
     /* decode input string */
     outp = out;
     state = ST_NORMAL;
@@ -1781,7 +1861,7 @@ static char *utf7_encode(char *in,int inlen,int *outlen)
       }
     }
     /* allocate output buffer */
-    if((out = ns_malloc((*outlen) + 1)) == NULL) return 0;
+    if((out = ns_malloc((unsigned)(*outlen) + 1)) == NULL) return 0;
     /* encode input string */
     outp = out;
     state = ST_NORMAL;
